@@ -64,17 +64,25 @@ html_print <- function(html, background = "white", viewer = getOption("viewer",
 #'
 #' @param html HTML content to print
 #' @param background Background color for web page
-#' @param file File to write content to
-#' @param libdir Directory to copy dependenies to
+#' @param file File path or connection. If a file path containing a
+#'   sub-directory, the sub-directory must already exist.
+#' @param libdir Directory to copy dependencies to
+#' @param lang Value of the `<html>` `lang` attribute
 #'
 #' @export
-save_html <- function(html, file, background = "white", libdir = "lib") {
+save_html <- function(html, file, background = "white", libdir = "lib", lang = "en") {
+  force(html)
+  force(background)
+  force(libdir)
 
   # ensure that the paths to dependencies are relative to the base
   # directory where the webpage is being built.
-  dir <- dirname(file)
-  oldwd <- setwd(dir)
-  on.exit(setwd(oldwd), add = TRUE)
+  if (is.character(file)) {
+    dir <- normalizePath(dirname(file), mustWork = TRUE)
+    file <- file.path(dir, basename(file))
+    owd <- setwd(dir)
+    on.exit(setwd(owd), add = TRUE)
+  }
 
   rendered <- renderTags(html)
 
@@ -84,21 +92,37 @@ save_html <- function(html, file, background = "white", libdir = "lib") {
     dep
   })
 
+  bodyBegin <- if (!isTRUE(grepl("<body\\b", rendered$html[1], ignore.case = TRUE))) {
+    "<body>"
+  }
+  bodyEnd <- if (!is.null(bodyBegin)) {
+    "</body>"
+  }
+
   # build the web-page
   html <- c("<!DOCTYPE html>",
-            "<html>",
+            sprintf('<html lang="%s">', lang),
             "<head>",
             "<meta charset=\"utf-8\"/>",
+            sprintf("<style>body{background-color:%s;}</style>", htmlEscape(background)),
             renderDependencies(deps, c("href", "file")),
             rendered$head,
             "</head>",
-            sprintf("<body style=\"background-color:%s;\">", htmlEscape(background)),
+            bodyBegin,
             rendered$html,
-            "</body>",
+            bodyEnd,
             "</html>")
 
+  if (is.character(file)) {
+    # Write to file in binary mode, so \r\n in input doesn't become \r\r\n
+    con <- base::file(file, open = "w+b")
+    on.exit(close(con), add = TRUE)
+  } else {
+    con <- file
+  }
+
   # write it
-  writeLines(html, file, useBytes = TRUE)
+  writeLines(html, con, useBytes = TRUE)
 }
 
 
