@@ -14,6 +14,11 @@ paste8 <- function(..., sep = " ", collapse = NULL) {
   do.call(paste, args)
 }
 
+# A special case of paste8 that employs paste0. Avoids the overhead of lapply.
+concat8 <- function(...) {
+  enc2utf8(paste0(...))
+}
+
 # Reusable function for registering a set of methods with S3 manually. The
 # methods argument is a list of character vectors, each of which has the form
 # c(package, genname, class).
@@ -68,17 +73,19 @@ depListToNamedDepList <- function(dependencies) {
 #' the latest version number is used.
 #'
 #' @param dependencies A list of \code{\link{htmlDependency}} objects.
+#' @param resolvePackageDir Whether to resolve the relative path to an absolute
+#'   path via \code{\link{system.file}} when the \code{package} attribute is
+#'   present in a dependency object.
 #' @return dependencies A list of \code{\link{htmlDependency}} objects with
 #'   redundancies removed.
 #'
 #' @export
-resolveDependencies <- function(dependencies) {
-  # Remove nulls
-  deps <- dependencies[!sapply(dependencies, is.null)]
+resolveDependencies <- function(dependencies, resolvePackageDir = TRUE) {
+  deps <- resolveFunctionalDependencies(dependencies)
 
   # Get names and numeric versions in vector/list form
-  depnames <- sapply(deps, `[[`, "name")
-  depvers <- numeric_version(sapply(deps, `[[`, "version"))
+  depnames <- vapply(deps, function(x) x$name, character(1))
+  depvers <- numeric_version(vapply(deps, function(x) x$version, character(1)))
 
   # Get latest version of each dependency. `unique` uses the first occurrence of
   # each dependency name, which is important for inter-dependent libraries.
@@ -88,7 +95,13 @@ resolveDependencies <- function(dependencies) {
     sorted <- order(ifelse(depnames == depname, TRUE, NA), depvers,
       na.last = NA, decreasing = TRUE)
     # The first element in the list is the one with the largest version.
-    deps[[sorted[[1]]]]
+    dep <- deps[[sorted[[1]]]]
+    if (resolvePackageDir && !is.null(dep$package)) {
+      dir <- dep$src$file
+      if (!is.null(dir)) dep$src$file <- system.file(dir, package = dep$package)
+      dep$package <- NULL
+    }
+    dep
   }))
 }
 
@@ -121,11 +134,15 @@ resolveDependencies <- function(dependencies) {
 #'
 #' @export
 subtractDependencies <- function(dependencies, remove, warnOnConflict = TRUE) {
-  depnames <- sapply(dependencies, `[[`, "name")
-  rmnames <- if (is.character(remove))
-    remove
-  else
-    sapply(remove, `[[`, "name")
+  dependencies <- resolveFunctionalDependencies(dependencies)
+  depnames <- vapply(dependencies, function(x) x$name, character(1))
+
+  if (is.character(remove)) {
+    rmnames <- remove
+  } else {
+    remove <- resolveFunctionalDependencies(remove)
+    rmnames <- vapply(remove, function(x) x$name, character(1))
+  }
 
   matches <- depnames %in% rmnames
   if (warnOnConflict && !is.character(remove)) {
@@ -150,9 +167,10 @@ dropNulls <- function(x) {
 }
 
 nullOrEmpty <- function(x) {
-  is.null(x) || length(x) == 0
+  length(x) == 0
 }
-# Given a vector or list, drop all the NULL items in it
+
+# Given a vector or list, drop all the NULL or length-0 items in it
 dropNullsOrEmpty <- function(x) {
   x[!vapply(x, nullOrEmpty, FUN.VALUE=logical(1))]
 }
@@ -167,7 +185,7 @@ print.shiny.tag <- function(x, browse = is.browsable(x), ...) {
   if (browse)
     html_print(x)
   else
-    print(as.character(x), ...)
+    print(HTML(as.character(x)), ...)
   invisible(x)
 }
 
@@ -180,7 +198,7 @@ format.shiny.tag <- function(x, ..., singletons = character(0), indent = 0) {
 
 #' @export
 as.character.shiny.tag <- function(x, ...) {
-  renderTags(x)$html
+  as.character(renderTags(x)$html)
 }
 
 #' @export
@@ -212,9 +230,9 @@ as.character.shiny.tag.list <- as.character.shiny.tag
 #' @export
 print.html <- function(x, ..., browse = is.browsable(x)) {
   if (browse)
-    html_print(HTML(x))
+    html_print(x)
   else
-    cat(x, "\n")
+    cat(x, "\n", sep = "")
   invisible(x)
 }
 
@@ -233,17 +251,48 @@ normalizeText <- function(text) {
 
 #' @name tag
 #' @rdname tag
+#' @import rlang
 #' @export
 tagList <- function(...) {
-  lst <- list(...)
+  lst <- dots_list(...)
   class(lst) <- c("shiny.tag.list", "list")
   return(lst)
 }
 
+#' Tag function
+#'
+#' Create 'lazily' rendered HTML [tags] (and/or [htmlDependencies()]).
+#'
+#' @param func a function with no arguments that returns HTML tags and/or
+#'   dependencies.
+#'
+#' @export
+#' @examples
+#'
+#' myDivDep <- tagFunction(function() {
+#'   if (isTRUE(getOption("useDep", TRUE))) {
+#'     htmlDependency(
+#'       name = "lazy-dependency",
+#'       version = "1.0", src = ""
+#'     )
+#'   }
+#' })
+#' myDiv <- attachDependencies(div(), myDivDep)
+#' renderTags(myDiv)
+#' withr::with_options(list(useDep = FALSE), renderTags(myDiv))
+#'
+tagFunction <- function(func) {
+  if (!is.function(func) || length(formals(func)) != 0) {
+    stop("`func` must be a function with no arguments")
+  }
+  structure(func, class = "shiny.tag.function")
+}
+
 #' @rdname tag
 #' @export
 tagAppendAttributes <- function(tag, ...) {
-  tag$attribs <- c(tag$attribs, list(...))
+  throw_if_tag_function(tag)
+  tag$attribs <- c(tag$attribs, dropNullsOrEmpty(dots_list(...)))
   tag
 }
 
@@ -251,6 +300,7 @@ tagAppendAttributes <- function(tag, ...) {
 #' @rdname tag
 #' @export
 tagHasAttribute <- function(tag, attr) {
+  throw_if_tag_function(tag)
   result <- attr %in% names(tag$attribs)
   result
 }
@@ -258,6 +308,7 @@ tagHasAttribute <- function(tag, attr) {
 #' @rdname tag
 #' @export
 tagGetAttribute <- function(tag, attr) {
+  throw_if_tag_function(tag)
   # Find out which positions in the attributes list correspond to the given attr
   attribs <- tag$attribs
   attrIdx <- which(attr == names(attribs))
@@ -276,6 +327,7 @@ tagGetAttribute <- function(tag, attr) {
 #' @rdname tag
 #' @export
 tagAppendChild <- function(tag, child) {
+  throw_if_tag_function(tag)
   tag$children[[length(tag$children)+1]] <- child
   tag
 }
@@ -283,17 +335,24 @@ tagAppendChild <- function(tag, child) {
 #' @rdname tag
 #' @export
 tagAppendChildren <- function(tag, ..., list = NULL) {
-  tag$children <- c(tag$children, c(list(...), list))
+  throw_if_tag_function(tag)
+  tag$children <- unname(c(tag$children, c(dots_list(...), list)))
   tag
 }
 
 #' @rdname tag
 #' @export
 tagSetChildren <- function(tag, ..., list = NULL) {
-  tag$children <- c(list(...), list)
+  throw_if_tag_function(tag)
+  tag$children <- unname(c(dots_list(...), list))
   tag
 }
 
+throw_if_tag_function <- function(tag) {
+  if (is_tag_function(tag))
+    stop("`tag` can not be a `tagFunction()`")
+}
+
 #' HTML Tag Object
 #'
 #' \code{tag()} creates an HTML tag definition. Note that all of the valid HTML5
@@ -313,6 +372,10 @@ tagSetChildren <- function(tag, ..., list = NULL) {
 #' @param ...  Unnamed items that comprise this list of tags.
 #' @param list An optional list of elements. Can be used with or instead of the
 #'   \code{...} items.
+#' @param .noWS Character vector used to omit some of the whitespace that would
+#'   normally be written around this tag. Valid options include \code{before},
+#'   \code{after}, \code{outside}, \code{after-begin}, and \code{before-end}.
+#'   Any number of these options can be specified.
 #' @return An HTML tag object that can be rendered as HTML using
 #'   \code{\link{as.character}()}.
 #' @export
@@ -327,34 +390,55 @@ tagSetChildren <- function(tag, ..., list = NULL) {
 #'           tags$h2("Header text"),
 #'           tags$p("Text here"))
 #' tagList(x)
-tag <- function(`_tag_name`, varArgs) {
+#'
+#' # suppress the whitespace between tags
+#' oneline <- tag("span",
+#'   tag("strong", "Super strong", .noWS="outside")
+#' )
+#' cat(as.character(oneline))
+tag <- function(`_tag_name`, varArgs, .noWS=NULL) {
+  validateNoWS(.noWS)
   # Get arg names; if not a named list, use vector of empty strings
   varArgsNames <- names(varArgs)
   if (is.null(varArgsNames))
     varArgsNames <- character(length=length(varArgs))
 
-  # Named arguments become attribs, dropping NULL values
+  # Named arguments become attribs, dropping NULL and length-0 values
   named_idx <- nzchar(varArgsNames)
-  attribs <- dropNulls(varArgs[named_idx])
+  attribs <- dropNullsOrEmpty(varArgs[named_idx])
 
   # Unnamed arguments are flattened and added as children.
   # Use unname() to remove the names attribute from the list, which would
   # consist of empty strings anyway.
   children <- unname(varArgs[!named_idx])
 
-  # Return tag data structure
-  structure(
-    list(name = `_tag_name`,
+  st <- list(name = `_tag_name`,
       attribs = attribs,
-      children = children),
-    class = "shiny.tag"
-  )
+      children = children)
+
+  # Conditionally include the .noWS element. We do this to avoid breaking the hashes
+  # of existing tags that weren't leveraging .noWS.
+  if (!is.null(.noWS)){
+    st$.noWS <- .noWS
+  }
+
+  # Return tag data structure
+  structure(st, class = "shiny.tag")
 }
 
 isTagList <- function(x) {
   is.list(x) && (inherits(x, "shiny.tag.list") || identical(class(x), "list"))
 }
 
+noWSOptions <- c("before", "after", "after-begin", "before-end", "outside", "inside")
+# Ensure that the provided `.noWS` string contains only valid options
+validateNoWS <- function(.noWS){
+  if (!all(.noWS %in% noWSOptions)){
+    stop("Invalid .noWS option(s) '", paste(.noWS, collapse="', '") ,"' specified.")
+  }
+}
+
+#' @include utils.R
 tagWrite <- function(tag, textWriter, indent=0, eol = "\n") {
 
   if (length(tag) == 0)
@@ -372,24 +456,47 @@ tagWrite <- function(tag, textWriter, indent=0, eol = "\n") {
 
   # compute indent text
   indentText <- paste(rep(" ", indent*2), collapse="")
+  textWriter$writeWS(indentText)
 
   # Check if it's just text (may either be plain-text or HTML)
   if (is.character(tag)) {
-    textWriter(indentText)
-    textWriter(normalizeText(tag))
-    textWriter(eol)
+    .noWS <- attr(tag, "noWS", exact = TRUE)
+    if ("before" %in% .noWS || "outside" %in% .noWS) {
+      textWriter$eatWS()
+    }
+    textWriter$write(normalizeText(tag))
+    if ("after" %in% .noWS || "outside" %in% .noWS) {
+      textWriter$eatWS()
+    }
+    textWriter$writeWS(eol)
     return (NULL)
   }
 
+  .noWS <- tag$.noWS
+
+  if ("before" %in% .noWS || "outside" %in% .noWS) {
+    textWriter$eatWS()
+  }
+
   # write tag name
-  textWriter(paste8(indentText, "<", tag$name, sep=""))
+  textWriter$write(concat8("<", tag$name))
 
   # Convert all attribs to chars explicitly; prevents us from messing up factors
   attribs <- lapply(tag$attribs, as.character)
   # concatenate attributes
   # split() is very slow, so avoid it if possible
-  if (anyDuplicated(names(attribs)))
-    attribs <- lapply(split(attribs, names(attribs)), paste, collapse = " ")
+  if (anyDuplicated(names(attribs))) {
+    attribs <- lapply(split(attribs, names(attribs)), function(x) {
+      na_idx <- is.na(x)
+      if (any(na_idx)) {
+        if (all(na_idx)) {
+          return(NA)
+        }
+        x <- x[!na_idx]
+      }
+      paste(x, collapse = " ")
+    })
+  }
 
   # write attributes
   for (attrib in names(attribs)) {
@@ -398,28 +505,34 @@ tagWrite <- function(tag, textWriter, indent=0, eol = "\n") {
       if (is.logical(attribValue))
         attribValue <- tolower(attribValue)
       text <- htmlEscape(attribValue, attribute=TRUE)
-      textWriter(paste8(" ", attrib,"=\"", text, "\"", sep=""))
+      textWriter$write(concat8(" ", attrib,"=\"", text, "\""))
     }
     else {
-      textWriter(paste8(" ", attrib, sep=""))
+      textWriter$write(concat8(" ", attrib))
     }
   }
 
   # write any children
   children <- dropNullsOrEmpty(flattenTags(tag$children))
   if (length(children) > 0) {
-    textWriter(">")
+    textWriter$write(">")
 
     # special case for a single child text node (skip newlines and indentation)
     if ((length(children) == 1) && is.character(children[[1]]) ) {
-      textWriter(paste8(normalizeText(children[[1]]), "</", tag$name, ">", eol,
-        sep=""))
+      textWriter$write(concat8(normalizeText(children[[1]]), "</", tag$name, ">"))
     }
     else {
-      textWriter("\n")
+      if ("after-begin" %in% .noWS || "inside" %in% .noWS) {
+        textWriter$eatWS()
+      }
+      textWriter$writeWS("\n")
       for (child in children)
         tagWrite(child, textWriter, nextIndent)
-      textWriter(paste8(indentText, "</", tag$name, ">", eol, sep=""))
+      textWriter$writeWS(indentText)
+      if ("before-end" %in% .noWS || "inside" %in% .noWS) {
+        textWriter$eatWS()
+      }
+      textWriter$write(concat8("</", tag$name, ">"))
     }
   }
   else {
@@ -428,12 +541,16 @@ tagWrite <- function(tag, textWriter, indent=0, eol = "\n") {
     if (tag$name %in% c("area", "base", "br", "col", "command", "embed", "hr",
       "img", "input", "keygen", "link", "meta", "param",
       "source", "track", "wbr")) {
-      textWriter(paste8("/>", eol, sep=""))
+      textWriter$write("/>")
     }
     else {
-      textWriter(paste8("></", tag$name, ">", eol, sep=""))
+      textWriter$write(concat8("></", tag$name, ">"))
     }
   }
+  if ("after" %in% .noWS || "outside" %in% .noWS) {
+    textWriter$eatWS()
+  }
+  textWriter$writeWS(eol)
 }
 
 #' Render tags into HTML
@@ -472,7 +589,7 @@ renderTags <- function(x, singletons = character(0), indent = 0) {
   # Do singleton and head processing before rendering
   singletonInfo <- takeSingletons(x, singletons)
   headInfo <- takeHeads(singletonInfo$ui)
-  deps <- resolveDependencies(findDependencies(singletonInfo$ui))
+  deps <- resolveDependencies(findDependencies(singletonInfo$ui, tagify = FALSE))
 
   headIndent <- if (is.numeric(indent)) indent + 1 else indent
   headHtml <- doRenderTags(headInfo$head, indent = headIndent)
@@ -491,31 +608,11 @@ renderTags <- function(x, singletons = character(0), indent = 0) {
 #' @rdname renderTags
 #' @export
 doRenderTags <- function(x, indent = 0) {
-  # The text that is written to this connWriter will be converted to
-  # UTF-8 using enc2utf8. The rendered output will always be UTF-8
-  # encoded.
-  #
-  # We use a file() here instead of textConnection() or paste/c to
-  # avoid the overhead of copying, which is huge for moderately
-  # large numbers of calls to connWriter(). Generally when you want
-  # to incrementally build up a long string out of immutable ones,
-  # you want to use a mutable/growable string buffer of some kind;
-  # since R doesn't have something like that (that I know of),
-  # file() is the next best thing.
-  conn <- file(open="w+b", encoding = "UTF-8")
-  connWriter <- function(text) {
-    text <- enc2utf8(text)
-    # This is actually writing UTF-8 bytes, not chars
-    writeBin(charToRaw(text), conn)
-  }
-  htmlResult <- tryCatch({
-    tagWrite(x, connWriter, indent)
-    flush(conn)
-    readLines(conn, encoding = "UTF-8")
-  },
-    finally = close(conn)
-  )
-  return(HTML(paste(htmlResult, collapse = "\n")))
+  textWriter <- WSTextWriter()
+  tagWrite(x, textWriter, indent)
+  # Strip off trailing \n (if present?)
+  textWriter$eatWS()
+  HTML(textWriter$readAll())
 }
 
 # Walk a tree of tag objects, rewriting objects according to func.
@@ -624,14 +721,16 @@ takeHeads <- function(ui) {
 #' Walks a hierarchy of tags looking for attached dependencies.
 #'
 #' @param tags A tag-like object to search for dependencies.
+#' @param tagify Whether to tagify the input before searching for dependencies.
 #'
 #' @return A list of \code{\link{htmlDependency}} objects.
 #'
 #' @export
-findDependencies <- function(tags) {
-  dep <- htmlDependencies(tags)
-  if (!is.null(dep) && inherits(dep, "html_dependency"))
-    dep <- list(dep)
+findDependencies <- function(tags, tagify = TRUE) {
+  if (isTRUE(tagify)) {
+    tags <- tagify(tags)
+  }
+  deps <- resolveFunctionalDependencies(htmlDependencies(tags))
   children <- if (is.list(tags)) {
     if (isTag(tags)) {
       tags$children
@@ -639,8 +738,35 @@ findDependencies <- function(tags) {
       tags
     }
   }
-  childDeps <- unlist(lapply(children, findDependencies), recursive = FALSE)
-  c(childDeps, if (!is.null(dep)) dep)
+  childDeps <- unlist(lapply(children, findDependencies, tagify = FALSE), recursive = FALSE, use.names = FALSE)
+  c(childDeps, deps)
+}
+
+
+#' Resolves any [tagFunction()]s inside a list of [htmlDependencies()]. To
+#' resolve [tagFunction()]s _and then_ remove redundant dependencies all at once,
+#' use [resolveDependencies()] (which calls this function internally).
+#' @noRd
+resolveFunctionalDependencies <- function(dependencies) {
+  if (!length(dependencies)) {
+    return(dependencies)
+  }
+  dependencies <- asDependencies(dependencies)
+  dependencies <- lapply(dependencies, function(dep) {
+    if (is_tag_function(dep)) {
+      dep <- dep()
+    }
+    if (isTag(dep) || inherits(dep, "shiny.tag.list")) {
+      warning(
+        "It appears attachDependencies() has been used to attach a tagFunction()",
+        "that returns a shiny.tag/shiny.tag.list, which is considered poor practice",
+        "since those tags will never actually get rendered", call. = FALSE
+      )
+      return(findDependencies(dep))
+    }
+    asDependencies(dep)
+  })
+  unlist(dependencies, recursive = FALSE, use.names = FALSE)
 }
 
 #' HTML Builder Functions
@@ -660,9 +786,19 @@ findDependencies <- function(tags) {
 #' @name builder
 #' @param ... Attributes and children of the element. Named arguments become
 #'   attributes, and positional arguments become children. Valid children are
-#'   tags, single-character character vectors (which become text nodes), and raw
-#'   HTML (see \code{\link{HTML}}). You can also pass lists that contain tags,
-#'   text nodes, and HTML.
+#'   tags, single-character character vectors (which become text nodes), raw
+#'   HTML (see \code{\link{HTML}}), and \code{html_dependency} objects. You can
+#'   also pass lists that contain tags, text nodes, or HTML. To use boolean
+#'   attributes, use a named argument with a \code{NA} value. (see example)
+#' @param .noWS A character vector used to omit some of the whitespace that
+#'   would normally be written around this tag. Valid options include
+#'   \code{before}, \code{after}, \code{outside}, \code{after-begin},
+#'   \code{before-end}, and \code{inside}. Any number of these options can be
+#'   specified.
+#' @references \itemize{
+#'    \item W3C html specification about boolean attributes
+#'    \url{https://www.w3.org/TR/html5/infrastructure.html#sec-boolean-attributes}
+#'  }
 #' @export tags
 #' @examples
 #' doc <- tags$html(
@@ -679,124 +815,48 @@ findDependencies <- function(tags) {
 #'   )
 #' )
 #' cat(as.character(doc))
+#'
+#' # create an html5 audio tag with controls.
+#' # controls is a boolean attributes
+#' audio_tag <- tags$audio(
+#'   controls = NA,
+#'   tags$source(
+#'     src = "myfile.wav",
+#'     type = "audio/wav"
+#'   )
+#' )
+#' cat(as.character(audio_tag))
+#'
+#' # suppress the whitespace between tags
+#' oneline <- tags$span(
+#'   tags$strong("I'm strong", .noWS="outside")
+#' )
+#' cat(as.character(oneline))
 NULL
 
+
+
+# Use `known_tags` from `known_tags.R`
+# Then remove `known_tags` once done creating tag functions
+#' @include known_tags.R
+names(known_tags) <- known_tags
+
 #' @rdname builder
 #' @format NULL
 #' @docType NULL
 #' @keywords NULL
-tags <- list(
-  a = function(...) tag("a", list(...)),
-  abbr = function(...) tag("abbr", list(...)),
-  address = function(...) tag("address", list(...)),
-  area = function(...) tag("area", list(...)),
-  article = function(...) tag("article", list(...)),
-  aside = function(...) tag("aside", list(...)),
-  audio = function(...) tag("audio", list(...)),
-  b = function(...) tag("b", list(...)),
-  base = function(...) tag("base", list(...)),
-  bdi = function(...) tag("bdi", list(...)),
-  bdo = function(...) tag("bdo", list(...)),
-  blockquote = function(...) tag("blockquote", list(...)),
-  body = function(...) tag("body", list(...)),
-  br = function(...) tag("br", list(...)),
-  button = function(...) tag("button", list(...)),
-  canvas = function(...) tag("canvas", list(...)),
-  caption = function(...) tag("caption", list(...)),
-  cite = function(...) tag("cite", list(...)),
-  code = function(...) tag("code", list(...)),
-  col = function(...) tag("col", list(...)),
-  colgroup = function(...) tag("colgroup", list(...)),
-  command = function(...) tag("command", list(...)),
-  data = function(...) tag("data", list(...)),
-  datalist = function(...) tag("datalist", list(...)),
-  dd = function(...) tag("dd", list(...)),
-  del = function(...) tag("del", list(...)),
-  details = function(...) tag("details", list(...)),
-  dfn = function(...) tag("dfn", list(...)),
-  div = function(...) tag("div", list(...)),
-  dl = function(...) tag("dl", list(...)),
-  dt = function(...) tag("dt", list(...)),
-  em = function(...) tag("em", list(...)),
-  embed = function(...) tag("embed", list(...)),
-  eventsource = function(...) tag("eventsource", list(...)),
-  fieldset = function(...) tag("fieldset", list(...)),
-  figcaption = function(...) tag("figcaption", list(...)),
-  figure = function(...) tag("figure", list(...)),
-  footer = function(...) tag("footer", list(...)),
-  form = function(...) tag("form", list(...)),
-  h1 = function(...) tag("h1", list(...)),
-  h2 = function(...) tag("h2", list(...)),
-  h3 = function(...) tag("h3", list(...)),
-  h4 = function(...) tag("h4", list(...)),
-  h5 = function(...) tag("h5", list(...)),
-  h6 = function(...) tag("h6", list(...)),
-  head = function(...) tag("head", list(...)),
-  header = function(...) tag("header", list(...)),
-  hgroup = function(...) tag("hgroup", list(...)),
-  hr = function(...) tag("hr", list(...)),
-  html = function(...) tag("html", list(...)),
-  i = function(...) tag("i", list(...)),
-  iframe = function(...) tag("iframe", list(...)),
-  img = function(...) tag("img", list(...)),
-  input = function(...) tag("input", list(...)),
-  ins = function(...) tag("ins", list(...)),
-  kbd = function(...) tag("kbd", list(...)),
-  keygen = function(...) tag("keygen", list(...)),
-  label = function(...) tag("label", list(...)),
-  legend = function(...) tag("legend", list(...)),
-  li = function(...) tag("li", list(...)),
-  link = function(...) tag("link", list(...)),
-  mark = function(...) tag("mark", list(...)),
-  map = function(...) tag("map", list(...)),
-  menu = function(...) tag("menu", list(...)),
-  meta = function(...) tag("meta", list(...)),
-  meter = function(...) tag("meter", list(...)),
-  nav = function(...) tag("nav", list(...)),
-  noscript = function(...) tag("noscript", list(...)),
-  object = function(...) tag("object", list(...)),
-  ol = function(...) tag("ol", list(...)),
-  optgroup = function(...) tag("optgroup", list(...)),
-  option = function(...) tag("option", list(...)),
-  output = function(...) tag("output", list(...)),
-  p = function(...) tag("p", list(...)),
-  param = function(...) tag("param", list(...)),
-  pre = function(...) tag("pre", list(...)),
-  progress = function(...) tag("progress", list(...)),
-  q = function(...) tag("q", list(...)),
-  ruby = function(...) tag("ruby", list(...)),
-  rp = function(...) tag("rp", list(...)),
-  rt = function(...) tag("rt", list(...)),
-  s = function(...) tag("s", list(...)),
-  samp = function(...) tag("samp", list(...)),
-  script = function(...) tag("script", list(...)),
-  section = function(...) tag("section", list(...)),
-  select = function(...) tag("select", list(...)),
-  small = function(...) tag("small", list(...)),
-  source = function(...) tag("source", list(...)),
-  span = function(...) tag("span", list(...)),
-  strong = function(...) tag("strong", list(...)),
-  style = function(...) tag("style", list(...)),
-  sub = function(...) tag("sub", list(...)),
-  summary = function(...) tag("summary", list(...)),
-  sup = function(...) tag("sup", list(...)),
-  table = function(...) tag("table", list(...)),
-  tbody = function(...) tag("tbody", list(...)),
-  td = function(...) tag("td", list(...)),
-  textarea = function(...) tag("textarea", list(...)),
-  tfoot = function(...) tag("tfoot", list(...)),
-  th = function(...) tag("th", list(...)),
-  thead = function(...) tag("thead", list(...)),
-  time = function(...) tag("time", list(...)),
-  title = function(...) tag("title", list(...)),
-  tr = function(...) tag("tr", list(...)),
-  track = function(...) tag("track", list(...)),
-  u = function(...) tag("u", list(...)),
-  ul = function(...) tag("ul", list(...)),
-  var = function(...) tag("var", list(...)),
-  video = function(...) tag("video", list(...)),
-  wbr = function(...) tag("wbr", list(...))
-)
+#' @import rlang
+tags <- lapply(known_tags, function(tagname) {
+  function(..., .noWS=NULL) {
+    validateNoWS(.noWS)
+    contents <- dots_list(...)
+    tag(tagname, contents, .noWS=.noWS)
+  }
+})
+
+# known_tags is no longer needed, so remove it.
+rm(known_tags)
+
 
 #' Mark Characters as HTML
 #'
@@ -806,6 +866,10 @@ tags <- list(
 #' @param text The text value to mark with HTML
 #' @param ... Any additional values to be converted to character and
 #'   concatenated together
+#' @param .noWS Character vector used to omit some of the whitespace that would
+#'   normally be written around this HTML. Valid options include \code{before},
+#'   \code{after}, and \code{outside} (equivalent to \code{before} and
+#'   \code{end}).
 #' @return The same value, but marked as HTML.
 #'
 #' @examples
@@ -813,10 +877,11 @@ tags <- list(
 #' cat(as.character(el))
 #'
 #' @export
-HTML <- function(text, ...) {
-  htmlText <- c(text, as.character(list(...)))
+HTML <- function(text, ..., .noWS = NULL) {
+  htmlText <- c(text, as.character(dots_list(...)))
   htmlText <- paste8(htmlText, collapse=" ")
   attr(htmlText, "html") <- TRUE
+  attr(htmlText, "noWS") <- .noWS
   class(htmlText) <- c("html", "character")
   htmlText
 }
@@ -898,7 +963,7 @@ flattenTags <- function(x) {
 #'
 #' An S3 method for converting arbitrary values to a value that can be used as
 #' the child of a tag or \code{tagList}. The default implementation simply calls
-#' \code{\link[base]{as.character}}.
+#' \code{\link{as.character}}.
 #'
 #' @param x Object to be converted.
 #' @param ... Any additional parameters.
@@ -910,10 +975,14 @@ as.tags <- function(x, ...) {
 
 #' @export
 as.tags.default <- function(x, ...) {
-  if (is.list(x) && !isTagList(x))
-    unclass(x)
-  else
+  # Plain (non-classed) lists will hit as.tags.list(), but lists with a class
+  # will get here. (tagLists will already have been handled by
+  # as.tags.shiny.tag.list)
+  if (is.list(x)) {
+    do.call(tagList, unclass(x))
+  } else {
     tagList(as.character(x))
+  }
 }
 
 #' @export
@@ -931,6 +1000,18 @@ as.tags.shiny.tag.list <- function(x, ...) {
   x
 }
 
+#' @export
+as.tags.shiny.tag.function <- function(x, ...) {
+  x()
+}
+
+#' @export
+as.tags.list <- function(x, ...) {
+  # Only non-classed lists will hit this method
+  # (classed lists will reach the default method)
+  do.call(tagList, x)
+}
+
 #' @export
 as.tags.character <- function(x, ...) {
   # For printing as.tags("<strong>") directly at console, without dropping any
@@ -938,6 +1019,11 @@ as.tags.character <- function(x, ...) {
   tagList(x)
 }
 
+#' @export
+as.tags.html_dependency <- function(x, ...) {
+  attachDependencies(tagList(), x)
+}
+
 #' Preserve HTML regions
 #'
 #' Use "magic" HTML comments to protect regions of HTML from being modified by
@@ -975,9 +1061,20 @@ as.tags.character <- function(x, ...) {
 #'
 #' @export
 htmlPreserve <- function(x) {
-  x <- paste(x, collapse = "\r\n")
+  raw = getOption("htmltools.preserve.raw", FALSE)
+  x <- paste(x, collapse = "\n")
   if (nzchar(x))
-    sprintf("<!--html_preserve-->%s<!--/html_preserve-->", x)
+    if (raw) {
+      # use fenced code block if there are embedded newlines
+      if (grepl("\n", x, fixed = TRUE))
+        sprintf("\n```{=html}\n%s\n```\n", x)
+      # otherwise use inline span
+      else
+        sprintf("`%s`{=html}", x)
+    }
+    else {
+      sprintf("<!--html_preserve-->%s<!--/html_preserve-->", x)
+    }
   else
     x
 }
@@ -1054,11 +1151,18 @@ extractPreserveChunks <- function(strval) {
     strval <- paste(strval, collapse = "\n")
 
   # matches contains the index of all the start and end markers
-  matches <- gregexpr(pattern, strval)[[1]]
-  lengths <- attr(matches, "match.length", TRUE)
+  startmatches <- gregexpr(startmarker, strval, fixed = TRUE)[[1]]
+  endmatches <- gregexpr(endmarker, strval, fixed = TRUE)[[1]]
+  matches <- c(startmatches, endmatches)
+  o <- order(matches)
+  matches <- matches[o]
+  lengths <- c(
+    attr(startmatches, "match.length", TRUE),
+    attr(endmatches, "match.length", TRUE)
+  )[o]
 
   # No markers? Just return.
-  if (matches[[1]] == -1)
+  if (unique(matches)[[1]] == -1)
     return(list(value = strval, chunks = character(0)))
 
   # If TRUE, it's a start; if FALSE, it's an end
@@ -1123,8 +1227,11 @@ extractPreserveChunks <- function(strval) {
 #' @rdname htmlPreserve
 #' @export
 restorePreserveChunks <- function(strval, chunks) {
+  strval <- enc2utf8(strval)
+  chunks <- enc2utf8(chunks)
   for (id in names(chunks))
     strval <- gsub(id, chunks[[id]], strval, fixed = TRUE, useBytes = TRUE)
+  Encoding(strval) <- 'UTF-8'
   strval
 }
 
@@ -1143,7 +1250,7 @@ NULL
 knit_print.shiny.tag <- function(x, ...) {
   x <- tagify(x)
   output <- surroundSingletons(x)
-  deps <- resolveDependencies(findDependencies(x))
+  deps <- resolveDependencies(findDependencies(x, tagify = FALSE), resolvePackageDir = FALSE)
   content <- takeHeads(output)
   head_content <- doRenderTags(tagList(content$head))
 
@@ -1160,7 +1267,7 @@ knit_print.shiny.tag <- function(x, ...) {
 #' @rdname knitr_methods
 #' @export
 knit_print.html <- function(x, ...) {
-  deps <- resolveDependencies(findDependencies(x))
+  deps <- resolveDependencies(findDependencies(x, tagify = FALSE))
   knitr::asis_output(htmlPreserve(as.character(x)),
     meta = if (length(deps)) list(deps))
 }
@@ -1173,71 +1280,71 @@ knit_print.shiny.tag.list <- knit_print.shiny.tag
 
 #' @rdname builder
 #' @export
-p <- function(...) tags$p(...)
+p <- tags$p
 
 #' @rdname builder
 #' @export
-h1 <- function(...) tags$h1(...)
+h1 <- tags$h1
 
 #' @rdname builder
 #' @export
-h2 <- function(...) tags$h2(...)
+h2 <- tags$h2
 
 #' @rdname builder
 #' @export
-h3 <- function(...) tags$h3(...)
+h3 <- tags$h3
 
 #' @rdname builder
 #' @export
-h4 <- function(...) tags$h4(...)
+h4 <- tags$h4
 
 #' @rdname builder
 #' @export
-h5 <- function(...) tags$h5(...)
+h5 <- tags$h5
 
 #' @rdname builder
 #' @export
-h6 <- function(...) tags$h6(...)
+h6 <- tags$h6
 
 #' @rdname builder
 #' @export
-a <- function(...) tags$a(...)
+a <- tags$a
 
 #' @rdname builder
 #' @export
-br <- function(...) tags$br(...)
+br <- tags$br
 
 #' @rdname builder
 #' @export
-div <- function(...) tags$div(...)
+div <- tags$div
 
 #' @rdname builder
 #' @export
-span <- function(...) tags$span(...)
+span <- tags$span
 
 #' @rdname builder
 #' @export
-pre <- function(...) tags$pre(...)
+pre <- tags$pre
 
 #' @rdname builder
 #' @export
-code <- function(...) tags$code(...)
+code <- tags$code
 
 #' @rdname builder
 #' @export
-img <- function(...) tags$img(...)
+img <- tags$img
 
 #' @rdname builder
 #' @export
-strong <- function(...) tags$strong(...)
+strong <- tags$strong
 
 #' @rdname builder
 #' @export
-em <- function(...) tags$em(...)
+em <- tags$em
 
 #' @rdname builder
 #' @export
-hr <- function(...) tags$hr(...)
+hr <- tags$hr
 
 #' Include Content From a File
 #'
@@ -1257,7 +1364,7 @@ hr <- function(...) tags$hr(...)
 #' @export
 includeHTML <- function(path) {
   lines <- readLines(path, warn=FALSE, encoding='UTF-8')
-  return(HTML(paste8(lines, collapse='\r\n')))
+  return(HTML(paste8(lines, collapse='\n')))
 }
 
 #' @note \code{includeText} escapes its contents, but does no other processing.
@@ -1270,7 +1377,7 @@ includeHTML <- function(path) {
 #' @export
 includeText <- function(path) {
   lines <- readLines(path, warn=FALSE, encoding='UTF-8')
-  return(paste8(lines, collapse='\r\n'))
+  return(paste8(lines, collapse='\n'))
 }
 
 #' @note The \code{includeMarkdown} function requires the \code{markdown}
@@ -1288,18 +1395,18 @@ includeMarkdown <- function(path) {
 #' @export
 includeCSS <- function(path, ...) {
   lines <- readLines(path, warn=FALSE, encoding='UTF-8')
-  args <- list(...)
+  args <- dots_list(...)
   if (is.null(args$type))
     args$type <- 'text/css'
   return(do.call(tags$style,
-    c(list(HTML(paste8(lines, collapse='\r\n'))), args)))
+    c(list(HTML(paste8(lines, collapse='\n'))), args)))
 }
 
 #' @rdname include
 #' @export
 includeScript <- function(path, ...) {
   lines <- readLines(path, warn=FALSE, encoding='UTF-8')
-  return(tags$script(HTML(paste8(lines, collapse='\r\n')), ...))
+  return(tags$script(HTML(paste8(lines, collapse='\n')), ...))
 }
 
 #' Include content only once
@@ -1334,11 +1441,14 @@ is.singleton <- function(x) {
 #' Single element numeric vectors are returned as a character vector with the
 #' number plus a suffix of \code{"px"}.
 #'
-#' Single element character vectors must be \code{"auto"} or \code{"inherit"},
-#' or a number. If the number has a suffix, it must be valid: \code{px},
-#' \code{\%}, \code{em}, \code{pt}, \code{in}, \code{cm}, \code{mm}, \code{ex},
-#' or \code{pc}. If the number has no suffix, the suffix \code{"px"} is
-#' appended.
+#' Single element character vectors must be \code{"auto"}, \code{"fit-content"}
+#' or \code{"inherit"}, a number, or a length calculated by the \code{"calc"}
+#' CSS function. If the number has a suffix, it must be valid: \code{px},
+#' \code{\%}, \code{ch}, \code{em}, \code{rem}, \code{pt}, \code{in}, \code{cm},
+#' \code{mm}, \code{ex}, \code{pc}, \code{vh}, \code{vw}, \code{vmin}, or
+#' \code{vmax}.
+#' If the number has no suffix, the suffix \code{"px"} is appended.
+#'
 #'
 #' Any other value will cause an error to be thrown.
 #'
@@ -1363,7 +1473,7 @@ validateCssUnit <- function(x) {
     x <- as.numeric(x)
 
   pattern <-
-    "^(auto|inherit|((\\.\\d+)|(\\d+(\\.\\d+)?))(%|in|cm|mm|em|ex|pt|pc|px))$"
+    "^(auto|inherit|fit-content|calc\\(.*\\)|((\\.\\d+)|(\\d+(\\.\\d+)?))(%|in|cm|mm|ch|em|ex|rem|pt|pc|px|vh|vw|vmin|vmax))$"
 
   if (is.character(x) &&
       !grepl(pattern, x)) {
@@ -1412,9 +1522,9 @@ validateCssUnit <- function(x) {
 #'
 #' @export
 css <- function(..., collapse_ = "") {
-  props <- list(...)
+  props <- dots_list(...)
   if (length(props) == 0) {
-    return("")
+    return(NULL)
   }
 
   if (is.null(names(props)) || any(names(props) == "")) {
@@ -1427,7 +1537,7 @@ css <- function(..., collapse_ = "") {
   # Drop null args
   props <- props[!sapply(props, empty)]
   if (length(props) == 0) {
-    return("")
+    return(NULL)
   }
 
   # Replace all '.' and '_' in property names to '-'
